"""
Tools for reading and manipulating SNP distribution data. For each individual and SNP location, it gives three probabilities: P(AA),P(AB),P(BB),
where A and B are the alleles. The probabilities should sum to 1.0. Missing data is represented by three numpy.NaN's.
"""

from pysnptools.distreader.distreader import DistReader
from pysnptools.distreader.distdata import DistData
from pysnptools.distreader.distnpz import DistNpz
from pysnptools.distreader.disthdf5 import DistHdf5
from pysnptools.distreader.distmemmap import DistMemMap
from pysnptools.distreader.bgen import Bgen
from pysnptools.distreader.distgen import DistGen
from pysnptools.distreader._distmergesids import _DistMergeSIDs

